/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package JamaPlus;

/**
 *
 * @author Bernard
 */
public class SparseMatrix {

    private int m, n, nnz;
    private int[] colpts, rowinds;
    private double[] vals;

    public SparseMatrix(int m, int n) { // null matrix
        this.m = m;
        this.n = n;
        nnz = 0;
        colpts = new int[n + 1];
        rowinds = new int[0];
        vals = new double[0];
    }

    private SparseMatrix(int m, int n, int nnz, int[] colpts, int[] rowinds, double[] vals) {
        this.m = m;
        this.n = n;
        this.nnz = nnz;
        this.colpts = colpts;
        this.rowinds = rowinds;
        this.vals = vals;
    }

    public SparseMatrix(Matrix mat) {  // convert a dense matrix
        m = mat.getRowDimension();
        n = mat.getColumnDimension();
        long nbval = mat.getNumberNonZeroValues();
        if (nbval > Integer.MAX_VALUE) throw new ArrayIndexOutOfBoundsException(
                    "Number of non-zero values exceeds 32 bits integer");
        nnz = (int) nbval;
        colpts = new int[n + 1];
        rowinds = new int[nnz];
        vals = new double[nnz];
        int nval = 0;
        for (int j = 0; j < n; j++) {
            colpts[j] = nval;
            for (int i = 0; i < m; i++) if (mat.get(i, j) != 0) {
                    rowinds[nval] = i;
                    vals[nval++] = mat.get(i, j);
                }
        }
        colpts[n] = nnz;
    }

 public SparseMatrix copy() {
     int[] colpts1 = new int[n+1];
     int[] rowinds1 = new int[nnz];
     double[] vals1 = new double[nnz];
     System.arraycopy(colpts, 0, colpts1, 0, n+1);
    System.arraycopy(rowinds, 0, rowinds1, 0, nnz);
    System.arraycopy(vals, 0, vals1, 0, nnz);
     return new SparseMatrix(m, n, nnz, colpts1, rowinds1, vals1);
 }
    public int getRowDimension() {
        return m;
    }

    public int getColumnDimension() {
        return n;
    }

    public double get(int i, int j) {
        try {
            for (int nval = colpts[j]; nval < colpts[j + 1]; nval++)
                if (rowinds[nval] == i) return vals[nval];
            return 0;
        } catch (ArrayIndexOutOfBoundsException ex) {
            throw new ArrayIndexOutOfBoundsException("Sparse matrix : "
                    + ex.getLocalizedMessage());
        }
    }

    public SparseMatrix transpose() {
        int[] colpts1 = new int[m + 1];
        int[] rowinds1 = new int[nnz];
        double[] vals1 = new double[nnz];
        for (int v = 0; v < nnz; v++) colpts1[rowinds[v]]++;  // nnz in each row
        colpts1[m] = nnz - colpts1[m - 1];
        for (int r = m - 1; r > 0; r--)
            colpts1[r] = colpts1[r + 1] - colpts1[r - 1]; // starting point of the previous row
        colpts1[0] = 0;
        int i = 0;
        for (int c = 0; c < n; c++) while (i < colpts[c + 1]) { 
                int j = colpts1[rowinds[i]+1]++;  // j : index of the value in the new vectors
                rowinds1[j] = c;
                vals1[j] = vals[i++];
            }
        return new SparseMatrix(n, m, nnz, colpts1, rowinds1, vals1);
    }
}
